from .exceptions import (
    InvalidMessageError,
    UnrecognizedProtocolVersion,
    WrongMessageTypeError,
)
from .message_types import Message
from .packer import DEFAULT_MESSAGE_PACKER, MessagePacker, pack, unpack

__all__ = (
    # main packing/unpacking interface
    "MessagePacker",
    "DEFAULT_MESSAGE_PACKER",
    "pack",
    "unpack",
    # common base for messages
    "Message",
    # errors
    "InvalidMessageError",
    "UnrecognizedProtocolVersion",
    "WrongMessageTypeError",
)



import abc
import typing as t

from ..tasks import TaskProtocol


class StorageException(Exception):
    def __init__(self, reason: str):
        self.reason = reason

    def __str__(self) -> str:
        return f"Storage request failed due to reason: {self.reason}"


class TaskStorage(abc.ABC):
    """
    Abstract class which defines the interface for task storage.
    """

    @abc.abstractmethod
    def store_result(self, task: TaskProtocol, result: str) -> None:
        """
        Store the result of a task.
        If the storage call succeeded, set reference to data in Task.result_reference
        If the storage failed, raises StorageException
        """

    @abc.abstractmethod
    def get_result(self, task: TaskProtocol) -> t.Optional[str]:
        """
        Get the result of a task.
        Returns a string if a result was found, and None otherwise.
        Raises a storage exception if retrieval failed
        """

    @abc.abstractmethod
    def store_payload(self, task: TaskProtocol, payload: str) -> None:
        """
        Store the payload of a task.
        If the storage call succeeded, set reference to data in Task.payload_reference
        If the storage failed, raises StorageException
        """

    @abc.abstractmethod
    def get_payload(self, task: TaskProtocol) -> t.Optional[str]:
        """
        Get the task payload.
        Returns a string if a payload was found, and None otherwise.
        Raises a storage exception if retrieval failed
        """



import contextlib
import logging
import os
import typing as t

try:
    import redis

    has_redis = True
except ImportError:
    has_redis = False

log = logging.getLogger(__name__)


def _check_has_redis() -> None:
    # defer this error until the caller tries to instantiate a connection or setup the
    # error log handler, so that imports work even without the 'redis' dependency
    if not has_redis:
        raise RuntimeError(
            """\
Cannot import globus_compute_common.redis if the 'redis' package is not available.
Either install it explicitly or install the 'redis' extra, as in

    pip install 'globus-compute-common[redis]'

"""
        )


def default_redis_connection_factory(
    redis_url: t.Optional[str] = None,
) -> "redis.Redis[str]":
    """
    Construct a Redis client for a given redis URL.
    If no URL is given, the COMPUTE_COMMON_REDIS_URL environment variable will be used.

    If no URL is given and the environment variable is not populated,

      redis://localhost:6379

    will be used as the default.
    """
    _check_has_redis()

    if redis_url is None:
        redis_url = os.getenv("COMPUTE_COMMON_REDIS_URL", "redis://localhost:6379")

    return redis.Redis.from_url(
        redis_url,
        decode_responses=True,
        health_check_interval=30,
    )


@contextlib.contextmanager
def redis_connection_error_logging(
    redis_client: "redis.Redis[t.Any]",
) -> t.Iterator[None]:
    _check_has_redis()

    try:
        yield
    except redis.exceptions.ConnectionError:
        log.exception(
            "ConnectionError while trying to communicate with redis, %s", redis_client
        )
        raise



import enum
import json
import typing as t


class ComputeRedisSerde:
    """
    A Serializer/Deserializer for data going into or out of Redis.

    The base implementation just stringifies data and does nothing when
    deserializing.
    """

    def serialize(self, value: t.Any) -> str:
        return str(value)

    def deserialize(self, value: str) -> t.Any:
        return value


class ComputeRedisIntSerde(ComputeRedisSerde):
    def deserialize(self, value: str) -> int:
        try:
            return int(value)
        except ValueError as e:
            raise ValueError(
                f"Invalid int value when loading from Redis: {value}"
            ) from e


class ComputeRedisFloatSerde(ComputeRedisSerde):
    def deserialize(self, value: str) -> float:
        try:
            return float(value)
        except ValueError as e:
            raise ValueError(
                f"Invalid float value when loading from Redis: {value}"
            ) from e


class ComputeRedisJSONSerde(ComputeRedisSerde):
    def serialize(self, value: t.Any) -> str:
        return json.dumps(value)

    def deserialize(self, value: str) -> t.Any:
        return json.loads(value)


class ComputeRedisEnumSerde(ComputeRedisSerde):
    def __init__(self, enum_class: t.Type[enum.Enum]) -> None:
        self.enum_class = enum_class

    def serialize(self, value: t.Any) -> str:
        return str(value.value)

    def deserialize(self, value: str) -> t.Any:
        return self.enum_class(value)


DEFAULT_SERDE = ComputeRedisSerde()
INT_SERDE = ComputeRedisIntSerde()
FLOAT_SERDE = ComputeRedisFloatSerde()
JSON_SERDE = ComputeRedisJSONSerde()



import enum


# subclass from str so that the enum can be JSON-encoded without adjustment
class TaskState(str, enum.Enum):
    RECEIVED = "received"  # on receiving a task web-side
    WAITING_FOR_EP = "waiting-for-ep"  # while waiting for ep to accept/be online
    WAITING_FOR_NODES = "waiting-for-nodes"  # jobs are pending at the scheduler
    WAITING_FOR_LAUNCH = "waiting-for-launch"
    EXEC_START = "execution-start"
    EXEC_END = "execution-end"
    RUNNING = "running"
    SUCCESS = "success"
    FAILED = "failed"
    RESULT_RECEIVED = "result-received"
    RESULT_ENQUEUED = "result-enqueued"
    AP_RECEIVED = "action-provider-received"
    AP_TASK_SUBMITTED = "action-provider-task-submitted"
    AP_TASKGROUP_RUNNING = "action-provider-task-group-in-progress"
    AP_TASKGROUP_COMPLETED = "action-provider-task-group-completed"
    AP_TASKGROUP_ERROR = "action-provider-task-group-error"


class InternalTaskState(str, enum.Enum):
    INCOMPLETE = "incomplete"
    COMPLETE = "complete"


class ActorName(str, enum.Enum):
    WORKER = "worker"
    MANAGER = "manager"
    INTERCHANGE = "interchange"
    ENDPOINT = "endpoint"
    RESULT_PROCESSOR = "result-processor"
    WEB_SERVICE = "web-service"
    ACTION_PROVIDER = "action-provider"



"""
The globus-compute-sdk passes a user-agent header string containing sdk version
information to the web service, which logs that information. This is a small,
centralized compatibility module for that behavior.
"""

import typing as t

GLOBUS_COMPUTE_VERSION_PREFIX = "globus-compute-sdk-"


def user_agent_substring(version_number: str) -> str:
    return GLOBUS_COMPUTE_VERSION_PREFIX + version_number


def parse_version_number(user_agent_string: str) -> t.Optional[str]:
    for part in user_agent_string.split("/"):
        if part.startswith(GLOBUS_COMPUTE_VERSION_PREFIX):
            return part.replace(GLOBUS_COMPUTE_VERSION_PREFIX, "")
    return None



from .connection import default_redis_connection_factory, redis_connection_error_logging
from .fields import HasRedisFields, HasRedisFieldsMeta, RedisField
from .pubsub import ComputeRedisPubSub
from .serde import (
    DEFAULT_SERDE,
    FLOAT_SERDE,
    INT_SERDE,
    JSON_SERDE,
    ComputeRedisEnumSerde,
    ComputeRedisFloatSerde,
    ComputeRedisIntSerde,
    ComputeRedisJSONSerde,
    ComputeRedisSerde,
)
from .task_queue import ComputeEndpointTaskQueue

__all__ = (
    "default_redis_connection_factory",
    "redis_connection_error_logging",
    "ComputeEndpointTaskQueue",
    "HasRedisFields",
    "HasRedisFieldsMeta",
    "RedisField",
    "ComputeRedisSerde",
    "ComputeRedisIntSerde",
    "ComputeRedisFloatSerde",
    "ComputeRedisJSONSerde",
    "DEFAULT_SERDE",
    "INT_SERDE",
    "FLOAT_SERDE",
    "JSON_SERDE",
    "ComputeRedisEnumSerde",
    "ComputeRedisPubSub",
)



from .base import StorageException, TaskStorage
from .default_storage import get_default_task_storage
from .redis import ImplicitRedisStorage
from .s3 import RedisS3Storage

__all__ = (
    "TaskStorage",
    "StorageException",
    "RedisS3Storage",
    "ImplicitRedisStorage",
    "get_default_task_storage",
)



import queue
import typing as t

from ..tasks import TaskProtocol, TaskState
from .connection import default_redis_connection_factory

if t.TYPE_CHECKING:
    import redis


class ComputeEndpointTaskQueue:
    def __init__(
        self, endpoint: str, *, redis_client: t.Optional["redis.Redis[t.Any]"] = None
    ) -> None:
        if redis_client is None:
            redis_client = default_redis_connection_factory()
        self.redis_client = redis_client
        self.endpoint = endpoint

    def __repr__(self) -> str:
        attr_str = f"endpoint={self.endpoint},redis_client={self.redis_client}"
        return f"ComputeEndpointTaskQueue({attr_str})"

    @property
    def queue_name(self) -> str:
        return f"task_{self.endpoint}_list"

    def enqueue(self, task: TaskProtocol) -> None:
        task.endpoint = self.endpoint
        task.status = TaskState.WAITING_FOR_EP
        self.redis_client.rpush(self.queue_name, task.task_id)

    def dequeue(self, *, timeout: int = 1) -> str:
        res = self.redis_client.blpop(self.queue_name, timeout=timeout)
        if not res:
            raise queue.Empty
        _queue_name, task_id = res
        return t.cast(str, task_id)



from __future__ import annotations

from .exceptions import UnrecognizedProtocolVersion
from .message_types import Message
from .protocol import MessagePackProtocol
from .protocol_versions.proto1 import MessagePackProtocolV1


class MessagePacker:
    IMPLEMENTATIONS: dict[int, MessagePackProtocol] = {
        1: MessagePackProtocolV1(),
    }

    def __init__(self, default_protocol_version: int = 1) -> None:
        self._default_protocol_version = default_protocol_version

    def detect_protocol_version(self, buf: bytes) -> int:
        """read the first byte of the buffer and decode it"""
        if not buf:
            raise ValueError("cannot detect_protocol_version on empty data")
        version_byte = buf[0:1]
        return int.from_bytes(version_byte, byteorder="big", signed=False)

    def pack(self, message: Message, *, protocol_version: int | None = None) -> bytes:
        if protocol_version is None:
            protocol_version = self._default_protocol_version
        impl = self.IMPLEMENTATIONS[protocol_version]
        return impl.pack(message)

    def unpack(self, buf: bytes) -> Message:
        protocol_version = self.detect_protocol_version(buf)
        try:
            impl = self.IMPLEMENTATIONS[protocol_version]
        except KeyError:
            raise UnrecognizedProtocolVersion(
                f"message had unknown protocol version {protocol_version}"
            )
        return impl.unpack(buf)


DEFAULT_MESSAGE_PACKER = MessagePacker()
pack = DEFAULT_MESSAGE_PACKER.pack
unpack = DEFAULT_MESSAGE_PACKER.unpack



import json
import typing as t

from .redis import (
    INT_SERDE,
    JSON_SERDE,
    ComputeRedisEnumSerde,
    HasRedisFieldsMeta,
    RedisField,
)
from .tasks import InternalTaskState, TaskProtocol, TaskState

try:
    import redis

    has_redis = True
except ImportError:
    has_redis = False


class RedisTask(TaskProtocol, metaclass=HasRedisFieldsMeta):
    """
    ORM-esque class to wrap access to properties of tasks.

    Creation:
      Create a new task by instantiating this class, e.g.
      >>> RedisTask(redis_client, "foo_id")

    Loading:
      Read a task from storage using the `load()` classmethod, e.g.
      >>> RedisTask.load(redis_client, "foo_id")

    Creation and Loading will each check `RedisTask.exists` when invoked. If the task_id
    already exists on creation, or does not exist on load, a ValueError will be raised.
    For example:
      >>> RedisTask.load(redis_client, "foo_id")
      Traceback (most recent call last):
        ...
      ValueError: Cannot load task foo_id: does not exist
      >>> RedisTask(redis_client, "foo_id")  # ok
      >>> RedisTask(redis_client, "foo_id")
      Traceback (most recent call last):
        ...
      ValueError: Conflict. Cannot create task foo_id: already exists

    This class provides various task fields as descriptors via RedisField, and is
    responsible for de/serializing various data from/to hstore in Redis.

    There are several elements of this pattern of use which need to be fixed. It is
    important to be aware of the following:
    - there is currently no use of Redis transactions, so nothing is atomic
    - Each time a field descriptor is accessed, it is read, returned, and discarded.
      Reading a field multiple times, even in a single python statement, is vulnerable
      to data races
    - no field requirements or validity are enforced -- reading a field can raise an
      error if bad data were written to Redis
    - each field is read individually, which can be inefficient and inconsistent (vs
      getall or setall semantics)
    """

    # 2 weeks in seconds
    DEFAULT_TTL: t.ClassVar[int] = 1209600

    # required fields
    # TODO: when `required=True` is supported in `RedisField`, set it for all of these
    status = t.cast(TaskState, RedisField(serde=ComputeRedisEnumSerde(TaskState)))
    internal_status = t.cast(
        InternalTaskState, RedisField(serde=ComputeRedisEnumSerde(InternalTaskState))
    )
    user_id = t.cast(int, RedisField(serde=INT_SERDE))
    function_id = t.cast(str, RedisField())
    container = t.cast(str, RedisField())
    task_group_id = t.cast(str, RedisField())
    queue_name = t.cast(str, RedisField())
    # end required fields

    endpoint_id = t.cast(t.Optional[str], RedisField())

    # FIXME: `payload` is a string which is currently being round-tripped through the
    # JSON_SERDE. However, we cannot remove the use of the serde until we are prepared
    # to handle the potential resulting errors. (Namely, that loading an old payload
    # would break)
    #
    # consider:
    #   >>> json.dumps("a")
    #   '"a"'
    #
    # therefore, we need to at least try `json.loads` on the payload (or do a breaking
    # change)
    #
    # alternatively, once `payload_reference` is populated on all tasks, we can use it
    # to include a bool flag for how the field should be deserialized. This would
    # require that the serde object itself have access to the payload_reference
    payload = t.cast(t.Optional[str], RedisField(serde=JSON_SERDE))
    payload_reference = t.cast(
        t.Optional[t.Dict[str, t.Any]], RedisField(serde=JSON_SERDE)
    )
    result = t.cast(t.Optional[str], RedisField())
    result_reference = t.cast(
        t.Optional[t.Dict[str, t.Any]], RedisField(serde=JSON_SERDE)
    )
    details = t.cast(t.Optional[t.Dict[str, t.Any]], RedisField(serde=JSON_SERDE))
    exception = t.cast(t.Optional[str], RedisField())
    completion_time = t.cast(t.Optional[str], RedisField())

    def __init__(
        self,
        redis_client: "redis.Redis[t.Any]",
        task_id: str,
        *,
        user_id: t.Optional[int] = None,
        function_id: t.Optional[str] = None,
        container: t.Optional[str] = None,
        payload: t.Optional[str] = None,
        payload_reference: t.Optional[t.Dict[str, t.Any]] = None,
        task_group_id: t.Optional[str] = None,
        queue_name: t.Optional[str] = None,
        endpoint_id: t.Optional[str] = None,
        details: t.Optional[t.Dict[str, t.Any]] = None,
    ):
        """
        If optional values are passed, then they will be written.

        Otherwise, they will fetched from any existing task entry.
        :param redis_client: Redis client for properties to get/set
        :param task_id: UUID of the task, as str
        :param user_id: ID of user to whom this task belongs
        :param function_id: UUID of the function for this task, as str
        :param container: UUID of container in which to run, as str
        :param payload: serialized function + input data
        :param task_group_id: UUID of task group that this task belongs to
        :param queue_name: name of AMQP queue where results will be sent
        :param endpoint_id: UUID of the endpoint the task was sent to
        """
        # non-RedisField attributes of a RedisTask
        self.hname = f"task_{task_id}"
        self.state_log_name = f"{self.hname}:state_log"
        self.redis_client = redis_client
        self.task_id = task_id

        # TODO: reject `RedisTask()` if the task_id already exists:
        #   if RedisTask.exists(redis_client, task_id): raise ...

        # if required attributes are not yet set, initialize them to their defaults
        if self.status is None:
            self.status = TaskState.WAITING_FOR_EP  # type: ignore[unreachable]
        if self.internal_status is None:
            self.internal_status = (  # type: ignore[unreachable]
                InternalTaskState.INCOMPLETE
            )

        # remaining RedisField attributes
        if user_id is not None:
            self.user_id = user_id
        if function_id is not None:
            self.function_id = function_id
        if container is not None:
            self.container = container
        if payload is not None:
            self.payload = payload
        if payload_reference is not None:
            self.payload_reference = payload_reference
        if task_group_id is not None:
            self.task_group_id = task_group_id
        if queue_name is not None:
            self.queue_name = queue_name
        if endpoint_id is not None:
            self.endpoint_id = endpoint_id
        if details is not None:
            self.details = details

        self.ttl = self.DEFAULT_TTL

    @property
    def ttl(self) -> int:
        return self.redis_client.ttl(self.hname)

    @ttl.setter
    def ttl(self, expiration: int) -> None:
        """Expires task after expiration time, if not already set."""
        ttl_val = self.redis_client.ttl(self.hname)
        if ttl_val < 0 or expiration < ttl_val:
            # expire was not already set
            self.redis_client.expire(self.hname, expiration)

    def delete(self) -> None:
        """Removes this task from Redis, to be used after the result is gotten"""
        self.redis_client.delete(self.hname)

    @classmethod
    def exists(cls, redis_client: "redis.Redis[t.Any]", task_id: str) -> bool:
        """Check if a given task_id exists in Redis"""
        return bool(redis_client.exists(f"task_{task_id}"))

    @classmethod
    def load(cls, redis_client: "redis.Redis[t.Any]", task_id: str) -> "RedisTask":
        """
        Load a task from storage. Raises a ValueError if the task is not found.
        """
        # TODO: This has a race condition. Encapsulate it in a transaction.
        if not cls.exists(redis_client, task_id):
            raise ValueError(f"Cannot load task {task_id}: does not exist")
        return cls(redis_client, task_id)

    @property
    def status_log(self) -> t.Iterable[t.Any]:
        return [
            json.loads(i) for i in self.redis_client.lrange(self.state_log_name, 0, -1)
        ]

    @status_log.setter
    def status_log(self, new_state: t.Any) -> None:
        self.redis_client.rpush(self.state_log_name, json.dumps(new_state))
        self.redis_client.expire(self.state_log_name, self.DEFAULT_TTL)



import logging
import queue
import typing as t

from ..tasks import TaskProtocol, TaskState
from .connection import default_redis_connection_factory

if t.TYPE_CHECKING:
    import redis

log = logging.getLogger(__name__)

_TASK_CHANNEL_PREFIX = "task_channel_"
_TASK_CHANNEL_PREFIX_LEN = len(_TASK_CHANNEL_PREFIX)
_TASK_QUEUE_PREFIX = "task_queue_"

_ALLOWED_MESSAGE_TYPES = ("pong", "message", "pmessage")


def _channel_name(endpoint_id: str) -> str:
    return f"{_TASK_CHANNEL_PREFIX}{endpoint_id}"


def _channel_name_to_endpoint_id(channel: str) -> str:
    return channel[_TASK_CHANNEL_PREFIX_LEN:]


def _queue_name(endpoint_id: str) -> str:
    return f"{_TASK_QUEUE_PREFIX}{endpoint_id}"


class ComputeRedisPubSub:
    """
    This class provides a layer over the Redis lib's `PubSub` functionality to
    push and pop messages into an endpoint_id-specific queue.
    Publishing and getting messages (plus subscribe/unsubscribe) are used to
    send messages via a pubsub channel in the normal case.

    If there is no recipient listening for a message, it is pushed into a queue instead.
    Subscribing pops all messages from the queue and puts them onto the pubsub channel.

    **IMPORTANT**

    Unsubscribing from a redis channel is not a synchronous operation. When
    unsubscribing, ensure clean teardown by calling ``get_final_messages()``.
    """

    def __init__(
        self, *, redis_client: t.Optional["redis.Redis[t.Any]"] = None
    ) -> None:
        if redis_client is None:
            redis_client = default_redis_connection_factory()
        self.redis_client = redis_client
        self.pubsub = self.redis_client.pubsub()

    def __repr__(self) -> str:
        return f"ComputeRedisPubSub(redis_client={self.redis_client})"

    @property
    def subscribed(self) -> bool:
        return bool(self.pubsub.subscribed)

    def put(self, endpoint_id: str, task: TaskProtocol) -> int:
        """
        Put the task ID into the channel for the endpoint.

        Returns the number of receipients who got the message.
        """
        # update the task object
        task.endpoint = endpoint_id
        task.status = TaskState.WAITING_FOR_EP

        # do the "main" publish step and record the number of recipients
        recipients = self.redis_client.publish(_channel_name(endpoint_id), task.task_id)

        # if there were no recipients for the published task_id, put it into
        # the task queue for that endpoint_id
        # when something subscribes to the endpoint channel, it can be
        # republished from there
        if recipients == 0:
            self.redis_client.rpush(_queue_name(endpoint_id), task.task_id)

        return recipients

    def republish_from_queue(self, endpoint_id: str) -> None:
        """
        Tasks pushed to Redis pubsub channels might have gone unreceived.
        When a new endpoint registers, it should republish tasks from it's queues
        to the pubsub channels.
        """
        # NOTE: this could block for an arbitrarily long period of time
        # if this becomes an issue, we can add a limit and batch the
        # resubmissions
        q = _queue_name(endpoint_id)
        channel = _channel_name(endpoint_id)

        # loop "until" trying to pop an item returns None
        queue_item = self.redis_client.blpop(q, timeout=1)
        while queue_item:
            task_list, task_id = queue_item

            # TODO: fix the fact that this does not check the number of
            # subscribers who received the message
            # because this is called from `subscribe()`, the normal case is for
            # this to be used by a client to get messages sent to itself for
            # later use in `get()` calls
            # if the publish step fails, it's not clear what that would mean
            # however, we could add an LPUSH call here to put the task_id back
            # into the queue
            self.redis_client.publish(channel, task_id)

            # pop the next item
            queue_item = self.redis_client.blpop(q, timeout=1)

    def subscribe(self, endpoint_id: str) -> None:
        channel = _channel_name(endpoint_id)
        log.info("subscribing to %s", channel)

        self.pubsub.subscribe(channel)
        self.republish_from_queue(endpoint_id)

    def unsubscribe(self, endpoint_id: str) -> None:
        channel = _channel_name(endpoint_id)
        log.info("unsubscribing from %s", channel)
        self.pubsub.unsubscribe(channel)

    def _get_message(self, timeout: float) -> t.Optional[t.Dict[str, t.Any]]:
        # skip any subscribe/unsubscribe messages, but do not use the
        # 'ignore_subscribe_messages' flag because it behaves by returning
        # `None` rather than advancing to the next message
        message = self.pubsub.get_message(timeout=timeout)
        while message is not None and message.get("type") not in _ALLOWED_MESSAGE_TYPES:
            message = self.pubsub.get_message(timeout=timeout)
        return message

    def get(self, *, timeout: int = 2) -> t.Tuple[str, str]:
        """
        :param timeout: wait time for getting a message, in milliseconds
        :type timeout: int
        """
        if not self.subscribed:
            raise queue.Empty

        message = self._get_message(timeout / 1000)
        if not message:
            raise queue.Empty("Channels empty")

        dest_endpoint = _channel_name_to_endpoint_id(message["channel"])
        task_id = message["data"]

        return dest_endpoint, task_id

    def _final_messages_generator(
        self, *, timeout: int
    ) -> t.Generator[t.Tuple[str, str], None, None]:
        while self.subscribed:
            try:
                yield self.get(timeout=timeout)
            # ignore empty responses during final consumption, since the whole
            # point is to consume until "the end of the queue"
            # but at that point, `self.subscribed` will become False
            except queue.Empty:
                pass

    def get_final_messages(
        self, *, timeout: int = 2
    ) -> t.Generator[t.Tuple[str, str], None, None]:
        """
        Yield back messages via ``get()`` for as long as the pubsub is marked
        as subscribed.
        """
        # type-ignore because this is not in the typeshed; TODO: get it added
        # to typeshed so that this type-checks
        num_pending_unsub = len(
            self.pubsub.pending_unsubscribe_channels  # type: ignore
        )
        if self.subscribed and (num_pending_unsub < len(self.pubsub.channels)):
            raise ValueError(
                "Cannot get final messages on this ComputeRedisPubSub. It has "
                "not been unsubscribed from all of its channels."
            )

        # Structure this as a function which returns a generator from another
        # call, don't use the yield syntax in this function body.
        # Why?
        # If we don't do this, the function gets compiled to bytecode for a
        # generator, and none of the body will be run until the first item is
        # fetched. That would mean the above ValueError does not get raised
        # when `get_final_messages()` is called, but when it is first used
        # instead. The difference is subtle, and may not matter for most
        # use-cases, but it prevents the possibility of someone thinking that
        # they've fully unsubscribed when they have not.
        return self._final_messages_generator(timeout=timeout)



"""
This module defines tools for helping to test funcx_common.
"""
try:
    import redis

    has_redis = True
except ImportError:
    has_redis = False


def _local_redis_reachable() -> bool:  # pragma: no cover
    if has_redis:
        try:
            redis.Redis("localhost", port=6379).ping()
            return True
        except redis.exceptions.ConnectionError:
            pass
    return False


# run this func only once to avoid slowing down the testsuite
LOCAL_REDIS_REACHABLE = _local_redis_reachable()



import typing as t

from ..tasks import TaskProtocol
from .base import TaskStorage


class ImplicitRedisStorage(TaskStorage):
    """
    This storage adapter stores task data to redis only.

    It is named "implicit" because it assumes that the task object itself is some form
    of RedisTask which will write to redis either on setattr or some save/commit step.
    """

    def store_result(
        self,
        task: TaskProtocol,
        result: str,
    ) -> None:
        task.result = result
        task.result_reference = {"storage_id": "redis"}

    def get_result(self, task: TaskProtocol) -> t.Optional[str]:
        if task.result:
            return task.result
        return None

    def store_payload(
        self,
        task: TaskProtocol,
        payload: str,
    ) -> None:
        task.payload = payload
        task.payload_reference = {"storage_id": "redis"}

    def get_payload(self, task: TaskProtocol) -> t.Optional[str]:
        if task.payload:
            return task.payload
        return None



from __future__ import annotations

import typing as t
import uuid

import pydantic

from .base import Message, meta
from .task_transition import TaskTransition


class ResultErrorDetails(pydantic.BaseModel):
    # a string for the error code
    code: str
    # the user is always supposed to see user_message somewhere
    user_message: str


@meta(message_type="result")
class Result(Message):
    task_id: uuid.UUID
    data: str
    details: t.Optional[t.Dict[t.Any, t.Any]]
    error_details: t.Optional[ResultErrorDetails]
    task_statuses: t.Optional[t.List[TaskTransition]]

    @property
    def is_error(self) -> bool:
        return self.error_details is not None



import typing as t
import uuid

from .base import Message, meta


@meta(message_type="containerimage")
class ContainerImage(Message):
    image_type: str  # e.g., (as currently envisioned), 'docker', 'singularity'
    location: str  # location or name required to pull the image (e.g., python:3.10)
    created_at: int
    modified_at: int
    build_status: t.Optional[str]
    build_stderr: t.Optional[str]


@meta(message_type="container")
class Container(Message):
    container_id: uuid.UUID
    name: str
    images: t.List[ContainerImage]
